package crystal;

import java.io.*;
import java.util.*;

import crystal.cfg.*;
import crystal.general.*;
import crystal.parser.*;
import crystal.analysis.callgraph.*;
import crystal.analysis.pointer.*;
import crystal.emitc.*;

/** The main class of the Crystal framework. */
public class Main {
	/** List of input C files to parse. */
	private static Vector<String> files = new Vector<String>();

	/** AST serialization: output file name. */
	private static String outputName = null;

	/** AST serialization: input file name. */
	private static String inputName = null;
	
	/** The main method.
	 *  The method is called by client analyses implemented in Crystal.
	 *  Clients must set up their options(in Options.* and elewehere), 
	 *  and then call this method to parse the files and build the IR. 
	 *  See an example in /projects/sample. */
	public static void main(String args[]) {
		/* Parse command line arguments. */
		parseArgs(args);
		
		try {
			long time = System.currentTimeMillis();
			if (inputName != null) {
				/* Load structures in from the input file. */
				System.out.print("Loading AST from input file \"" +
								 inputName + "\" ...");
				FileInputStream fis = new FileInputStream(inputName);
				ObjectInputStream inputStream = new ObjectInputStream(fis);
				deserialize(inputStream);
				System.out.printf("done. (%.2f s)%n", getSeconds(time));
			}
			else {
				/* Parse the input C files. */
				Actions.startParsing();
				int i = 1;
				for(String file : files) {
					System.out.printf("(%02d) Parsing file: \"%s\"", i++, file);
					long perfiletime = System.currentTimeMillis ();
					new parser(file).parse();
					System.out.printf(" (%.2f s)%n", getSeconds(perfiletime));
					System.out.flush();
				}
				Actions.doneParsing();
				System.out.printf("Parsed %d file%s, %d functions " + 
								  "successfully. (%.2f s)%n",
							 	  files.size(), files.size() > 1 ? "s" : "",
							 	  Symtab.getFunctions().size(),
							 	  getSeconds(time));
				if (Options.VERBOSITY >= 2) 
					printMemoryUsage();
			}

			if (outputName != null) {
				/* Write structures to the output file. */
				System.out.printf("Writing AST to output file \"" +
								  outputName + "\" ...");
				time = System.currentTimeMillis();
				FileOutputStream fos = new FileOutputStream(outputName);
				ObjectOutputStream outputStream = new ObjectOutputStream(fos);
				serialize(outputStream);
				System.out.printf("done. (%.2f s)%n", getSeconds(time));
				return;
			}

			
			if (Options.CFG) {
				System.out.print("Build CFGs ... ");
				time = System.currentTimeMillis();
				ScopeCollapse.collapseScopes();
				CFGBuilder.buildCFGs();
				System.out.printf("done. (%.2f s)%n", getSeconds(time));
				if (Options.VERBOSITY >= 2) 
					printMemoryUsage();
				
				if (Options.EMIT_DOT) { 
					System.out.println("Print CFGs to dot files");
					time = System.currentTimeMillis();
					CFGDump.dumper.printAll();
					System.out.printf("done. (%.2f s)%n", getSeconds(time));
				}
				
				if (Options.ALLOC_WRAPPER_DETECTION) {
					System.out.print("Alloc wrappers ... ");
					time = System.currentTimeMillis();
					Set<Function> wrap = AllocWrapperDetector.run();
					System.out.printf("done. (%.2f s)%n", getSeconds(time));
					if (Options.VERBOSITY >= 2) {
						System.out.println("   Detected alloc wrappers: " + wrap);
						printMemoryUsage();
					}
				}

				if (Options.POINTER_ANALYSIS) {
					System.out.print("Pointer analysis ... ");
					time = System.currentTimeMillis();
					PointerAnalysis.run();
					System.out.printf("done. (%.2f s)%n", getSeconds(time));
					if (Options.VERBOSITY >= 2) 
						printMemoryUsage();
				}
			}
			
			if (Options.CALL_GRAPH) {
				System.out.print("Call graph ... ");
				time = System.currentTimeMillis();
				CallGraph.construct();
				System.out.printf("done. (%.2f s)%n", getSeconds(time));
				if (Options.VERBOSITY >= 2) 
					printMemoryUsage();
			}

			if (Options.EMIT_C) {
				System.out.printf("Emit C code from the %s", 
								  (Options.CFG ? "CFG" : "AST"));
				time = System.currentTimeMillis();
				EmitC.run(true, false);
				System.out.printf(" (%.2f s)%n", getSeconds(time));
			}
		} 
        catch (SemanticError e) {
        	System.out.flush();
        	System.err.println("\nError: \"" + e.file + "\", line " +
        			e.line + " : " + e.message);
        	System.exit(1);
        } 
        catch (Exception e) {
        	System.out.flush();
        	System.err.println(e.getMessage());
        	e.printStackTrace();
        	System.exit(1);
        }
    }

	/** Returns seconds since the given start time. */
	static float getSeconds(long time) {
		time = System.currentTimeMillis() - time;
		return ((float)(time/1000.0));
	}
	
	public static void printMemoryUsage() {
		long tot = Runtime.getRuntime().totalMemory();
		long free = Runtime.getRuntime().freeMemory();
		double mb = 1024.0 * 1024.0;
		System.out.printf("Memory usage: %.2f Mb used, %.2f Mb total.%n", 
				(tot-free)/mb, tot/mb);
	}
	
	private static void serialize(ObjectOutputStream outputStream) 
		throws IOException 
	{
		outputStream.writeObject(Symtab.getGlobalScope());
		AggregateType.serializeStatic(outputStream);
	}

	private static void deserialize(ObjectInputStream inputStream) 
		throws IOException, ClassNotFoundException 
	{
		Symtab.setGlobalScope((HashSymtab)inputStream.readObject());
		AggregateType.deserializeStatic(inputStream);
	}

	/** Parse the comand line arguments. */
	static void parseArgs(String[] args) {
		if (args.length == 0)
			usage();
		
		for(int i = 0; i < args.length; i++) {
			String arg = args[i];
			if (arg.equals("-ast")) {
				Options.CFG = false;
			} 
			else if (arg.equals("-cfg")) {
				Options.CFG = true;
			} 
			else if (arg.equals("-emit-dot")) {
				Options.CFG = true;
				Options.EMIT_DOT = true;
			} 
			else if (arg.equals("-emit-c")) {
				Options.EMIT_C = true;
			} 
			else if (arg.startsWith("-")) {
				error(String.format("option %s not supported.", arg)); 
			} 
			else if (arg.equals("-out=")) {
				outputName = value(args[i]);
				Options.CFG = false;
			} 
			else if (arg.equals("-in=")) {
				inputName = value(args[i]);
			}
			else if (arg.startsWith("-v=")) {
				try {
					Options.VERBOSITY = Integer.parseInt(value(args[i]));
				} catch (Exception e) {
					error("option -v requires a number.\n");
				}
			} 
			else if (arg.startsWith("-alloc=")) {
				for (String name : value(args[i]).split(","))
					Options.allocNames.add(name);
			} 
			else if (arg.startsWith("-free=")) {
				for (String name : value(args[i]).split(","))
					Options.freeNames.add(name);
			} 
			else  {
				files.add(arg);
			}
		}
		
		if (Options.POINTER_ANALYSIS && !Options.CFG)
			error("pointer analysis requires CFG construction.");
		
		if (Options.ALLOC_WRAPPER_DETECTION && !Options.CFG)
			error("Alloc wrapper detection requires CFG construction.");
		
		if (Options.CALL_GRAPH && !Options.CFG)
			error("the call graph requires CFG construction.");
		
		if (Options.EMIT_DOT && !Options.CFG)
			error("dot files require CFG construction.");
		
		if (inputName != null && files.size() > 0) 
			error("cannot both read input and parse files.");

		if (inputName == null && files.size() == 0) 
			error("no files specified.");
	}
	
	private static String value(String s) {
		return s.substring(s.indexOf("=")+1);
	}

	/** Print error message and exit. */
	private static void error(String msg) {
		System.out.printf("%s : %s\n", Options.PROGNAME, msg);
		System.exit(1);
	}
	
	/** Print the command line usage. */
	static void usage() {
		System.out.printf ("Usage: %s [options] [preprocessed C files]\n", 
				Options.PROGNAME);
		
		String options = 
			"Options:\n"  +
			"    -ast        build the AST \n" +
			"    -cfg        build the CFG (default) \n" +
			"    -emit-c     emit output C code\n" +
			"    -emit-dot   print the CFGs to dot files\n";
		
		System.out.print(options);
		System.exit(0);
	}
}
