package crystal.cfg;

import crystal.ast.*;
import crystal.core.*;

import java.util.*;

/**
 * This represent a conditional statement, which evaluates some expression,
 *  and based on that, goes someplace.
 */
public class CFGSwitchNode extends CFGNode {

	/** Successor edges */
	List<CFGNode> outEdges;
	
	/** The test expression */
	CoreExpr test;
	Expression testSrc;

	/** Edge conditions, showing when the branch is taken. 
	 *  The default case has a null expression; and must be the last
	 *  in the vector of cases. */
	CoreExpr[] labels;
	List<Expression> labelsSrc;
	
	CFGSwitchNode(Expression e) {
		labelsSrc = new ArrayList<Expression>();
		testSrc = e;
		outEdges = new ArrayList<CFGNode>();
	}
	
	public boolean isSwitch() {
		return true;
	}

	public void buildCoreExpressions() {
		test = CoreExpr.fromAST(testSrc);
		labels = new CoreExpr[labelsSrc.size()];
		for (int i = 0; i < labelsSrc.size(); i++) {
			Expression lab = labelsSrc.get(i);
			labels[i] = lab  == null ? null : CoreExpr.fromAST(lab);
		}
	}
	
	public int numSuccs() {
		return outEdges.size();
	}
	
	public CFGNode getSucc(int k) {
		return outEdges.get(k);
	}

	void setSucc(int k, CFGNode n) {
		outEdges.set(k, n);
	}

	boolean hasDefault() {
		int cases = outEdges.size(); 
		return  cases > 0 && labelsSrc.get(cases-1) == null;
	}
	
	void addEdge(CFGNode dest, Expression label) {
		if (label != null) {
			labelsSrc.add(0, label);
			outEdges.add(0, dest);
		}
		else {
			labelsSrc.add(null);
			outEdges.add(dest);
		}
	}
	
	public int numEdges() {
		return outEdges.size();
	}
	
	public CFGNode getEdge(int i) {
		return outEdges.get(i);
	}

	public CoreExpr getLabel(int i) {
		return labels[i];
	}
	
	public Expression getLabelSrc(int i) {
		return labelsSrc.get(i);
	}
	
	public CoreExpr getTest() {
		return test;
	}

	public Expression getTestSrc() {
		return testSrc;
	}

	public CoreExpr getCoreKid(int i) {
		switch (i) {
		case 0: return test;
		default: return null;
		}
	}

	public int numCoreKids() {
		return 1;
	}

	public String toString() {
		return "switch " + ASTPrinter.toString(testSrc, true);
	}

	public CFGNode getDefaultSucc() {
		for (int i = 0; i < outEdges.size(); i++)
			if (labelsSrc.get(i) == null)
				return outEdges.get(i);

		return null;
	}

	public void addThenElse(CFGNode nthen, CFGNode nelse) {
		if (!testSrc.equals(NumberExpr.one)) {
			labelsSrc.add(0, NumberExpr.zero);
			outEdges.add(0, nelse);
		}
		labelsSrc.add(null);
		outEdges.add(nthen);
	}
	
    public boolean isIfThenElse() {
		if (numEdges() == 2 &&
			labelsSrc.get(0) instanceof NumberExpr &&
			labelsSrc.get(1) == CaseLabelStat.DEFAULT) 
		{
			NumberExpr ne = (NumberExpr)labelsSrc.get(0);
			return ne == NumberExpr.zero;
		}
		return false;
    }
    

}
