package crystal.general;

import java.io.Serializable;

public class Symbol implements Serializable {
	private static final long serialVersionUID = 3258130262767711032L;

	/** If true, prefix symbol names with their numeric identifiers.
	 *  For debugging purposes, to distinguish symbols with the same name. */
	public static boolean DEBUG_SYM = false;
	
	/** Maximum number of tries when generating fresh names 
	 *  during alpha-renaming */
	public static int FRESH_NAME_MAX_COUNTER = 10000;

	/** The counter for numeric identifiers. */
	private static long cid = 1;
	
	/** The numeric identifier of the symbol. */
    public long id;
    
    /** A special symbol denoting varargs. */
    public static final Symbol varargs;
    
    static { 
        varargs = newTemp("...", Type.VarArgs);
    }

    /** Symbol name */
    private String name;
    
    /** Symbol type */
    Type type;

    /** The source location where the symbol has been declared. */
    Loc loc;

    /** The symbol kind (VAR, FUN, TYPE, STRUCT, UNION, ENUM, LABEL) */
    byte kind;

    /** The symbol status (DEFINED, COMMON, UNDEFINED, NONE) */
    byte status;
        
    /** Storage information. */
    byte storage;
    
    /** Other information */
    byte info;
    
    
    public static Symbol newTemp(String s, Type t) {
    	return new Symbol(s, t, Flags.VAR, Flags.DEFINED, 0, Loc.UNK);
    }
    
    public static Symbol newVar(String s, Type t, Loc l) {
    	return new Symbol(s, t, Flags.VAR, Flags.DEFINED, 0, l);
    }
    
    public static Symbol newStruct(String s, byte k) {
    	return new Symbol(s, null, k, Flags.UNDEFINED, 0, Loc.UNK);
    }
    
    public Symbol(String n, Type t, byte k, byte s, int r, Loc l) {
    	name = n;
        type = t;
        kind = k;
        status = s;
        loc = l;
        storage = (byte)(r & Flags.STORAGE);
        id = cid++;
    }

    public String toString() {
    	return name;
    }

    public String getName() {
    	return name;
    }

    public void setName(String n) {
    	name = n;
    }

    public String getIdName() {
    	if (DEBUG_SYM) 
    		return "_" + id + "_" + name;
    	else
    		return name;
    }
    

	public String getNameAndFile() {
		return "\"" + name + "\" (in " + loc + ")";
	}

	public String getQuotName() {
    	return "\"" + name + "\"";
    }
    
    public Type getType() {
        return type;
    }
    
    public byte getKind() {
        return kind;
    }
    
    public byte getStatus() {
        return status;
    }

    public byte getStorage() {
        return storage;
    }
    
    public void setLoc(Loc l) {
    	loc = l;
    }
    
    public Loc getLoc() {
    	return loc;
    }
    
    public void setFreshName(Symtab tab) {
    	for (int i = 1; i < FRESH_NAME_MAX_COUNTER; i++) 
    		if (tab.lookupSymbol(kind, name + "_" + i) == null) {
    			name += "_" + i;
    			return;
    		}
    	throw new RuntimeException("alpha renaming failed for: " + name);
    }
    
    public void setType(Type type) {
        this.type = type;
    }
    
    public void setStatus(byte status) {
        this.status = status;
    }
    
    public void addStorage(int moreStorage) {
        this.storage |= (byte)moreStorage;
    }
    
    public boolean isTag() {
    	return kind == Flags.STRUCT || 
    	       kind == Flags.UNION || 
    		   kind == Flags.ENUM;
    }

    public boolean isDefined() {
    	return status == Flags.DEFINED;
    }

    public boolean isUndefined() {
    	return status == Flags.UNDEFINED;
    }
    
    public boolean isFunction() {
    	return false;
    }
    
    public boolean isStatic() {
    	return (storage & Flags.STATIC) != 0;
    }
    
    public void setStatic() {
    	storage |= Flags.STATIC;
    }
    
	public boolean isAutomatic() {
		return kind == Flags.VAR && !isStatic() && !isShadow();
	}

	public boolean isInline() {
		return false;
	}

	public void setAddressTaken() {
        this.info |= Flags.ADDR_TAKEN;
    }
    
    public boolean hasAddressTaken() {
    	return (this.info & Flags.ADDR_TAKEN) != 0; 
    }
    
    public void setFirstField() {
    	this.info |= Flags.FIRST_FIELD;
    }
    
    public boolean isFirstField() {
    	return (this.info & Flags.FIRST_FIELD) != 0; 
    }
    
    public void setAnonymous() {
    	this.info |= Flags.ANONYM;
    }
    
    public boolean isAnonymous() {
    	return (this.info & Flags.ANONYM) != 0; 
    }
    
    public void setShadow() {
    	this.info |= Flags.SHADOW;
    }
    
    public boolean isShadow() {
    	return (this.info & Flags.SHADOW) != 0; 
    }
    
    public void setAccessed() {
    	this.info |= Flags.ACCESSED;
    }
    
    public boolean isAccessed() {
    	return (this.info & Flags.ACCESSED) != 0; 
    }

    public void setParameter() {
    	this.info |= Flags.PARAMETER;
    }
    
    public boolean isParameter() {
    	return (this.info & Flags.PARAMETER) != 0; 
    }
    
	public void print(String s) {
    	System.out.print(s);
        System.out.print("(" + ((char)status) + ") ");
        if (isParameter())
        	System.out.print(String.format("param "));
        else
        	System.out.print(String.format("%-6s", Flags.kindToString(kind)));
        System.out.print(" : " + String.format("%-12s", getIdName()));
        System.out.print(" : " + (type == null? "[null]" : type.toString()));
        System.out.println();
    }
}

