package crystal.parser;

import java.util.*;

import crystal.ast.*;
import crystal.general.*;
import crystal.util.Order;

public class Actions {
	private static final boolean SANITY_CHECKS = true;
	
	/* A global symbol table for all files parsed so far. */
	private static HashSymtab allglobal = null;
	
	/* The block where initializations must go. When we're at the file scope,
	 * this block is the body of the file-wide initialization function */
	private static Stack<BlockStat> initBlockStack = new Stack<BlockStat>();

    /* The main and startup functions */
    private static Function mainfun = null;
    private static List<Function> inits = new ArrayList<Function>();

    /* Indicates we're parsing old-style parameters. */
    private static Symtab oldStyleParams = null;
    
    /* A stack of symbols to be initialized. */
    protected static Stack<Symbol>initStack = new Stack<Symbol>();

    /* A stack of specifiers */
    protected static Stack<Specifiers> specifiersStack = new Stack<Specifiers>();
    
    /* The stack of scopes where struct/union/enum symbols are added.
     * Different than the Symtab's scope stack when we enter nested
     * struct scopes. When the top of the stack is null, we're in a 
     * function param scope, and no new struct declarations are allowed. */
    private static Stack<Symtab> structScopeStack = new Stack<Symtab>();
    
    /* counter for default struct names */
    private static int c = 1;

    /** Process a declaration and enter the symbol it declares 
     *  in the appropriate symbol table. Uses the specifiers on
     *  top of the specifiers stack. Returns the declared symbol.
     *  @param popit: if true, pop the specifier stack
     *  @param init: if true, the declaration initializes the symbol */
    static Symbol processDeclaration(Declarator d, boolean popit, boolean init) 
    		    throws SemanticError 
    {
        Specifiers s = popit ? specifiersStack.pop()
                             : specifiersStack.peek();

        Symbol sym = d.getSymbol(s, Flags.UNDEFINED);

        /* function param void declarations must be alone. */
        if (sym.getType() == Type.Void) {
        	Symtab stab = Symtab.getCurrentScope();
        	if (stab instanceof FunSymtab) {
            	FunSymtab fstab = (FunSymtab) stab;
            	if (fstab.hasVoidArgs() || fstab.getParams().size() > 0)
            		error("void argument must be alone");            	
            	fstab.setVoidArgs();
            	return sym;
        	}
        }

        boolean oldStyle = oldStyleParams == Symtab.getCurrentScope();
    	Symtab crt = Symtab.getCurrentScope();
    	Symtab global = Symtab.getGlobalScope();
    	boolean globalScope = crt.getParent() == global;
        boolean declaredExtern = s.hasFlags(Flags.EXTERN); 
        boolean declaredStatic = s.hasFlags(Flags.STATIC);
        boolean declaredTypedef = s.hasFlags(Flags.TYPEDEF);
    	boolean defaultExtern = 
    		!declaredTypedef && !declaredStatic &&
    		(sym.isFunction() || crt instanceof FileSymtab);
        boolean isExtern = declaredExtern || defaultExtern; 
        
    	if (declaredExtern && declaredStatic)
    		error("declaration is both extern and static");
            	
    	if (declaredTypedef && (declaredStatic || declaredExtern))
    		error("typedef cannot be extern or static");    		
    	
    	if (init && declaredTypedef)
    		error("typedef cannot be initialized");    		
    	
    	if (init && declaredExtern)
    		error("extern symbol cannot be initialized");    		
    	
        /* calculate symbol status */
        byte status;
        if(declaredExtern || sym.isFunction())
        	status = Flags.UNDEFINED;
        else if (declaredTypedef || init || !globalScope)
        	status = Flags.DEFINED;
        else
        	status = Flags.COMMON;
        sym.setStatus(status);
        
        /* the symtab where the symbol will be inserted. */
        Symtab thetab = crt;

        if (isExtern && !oldStyle) {
        	checkAll(sym, crt, global);
        	thetab = global;
        }
        
        Symbol oldsym = thetab.lookupSymbol(sym);

        if (oldsym == null) {
        	if (oldStyle) 
        		error("invalid old-style parameter " + sym.getQuotName());
            thetab.insertSymLocal(sym);
            return sym;        	
        } else {
        	if (oldStyle) 
        		oldsym.setType(sym.getType());
        	else 
        		multipleDeclarations(sym, oldsym);
        	
        	return oldsym;	
        }
    }

    private static void multipleDeclarations(Symbol newsym, Symbol oldsym) 
    					throws SemanticError 
    {
		if (newsym.getKind() != oldsym.getKind())
			error(newsym.getQuotName() + 
					" redeclared as a different kind of symbol");

		byte status = newsym.getStatus();
		boolean def = status == Flags.DEFINED;
		byte oldstatus = oldsym.getStatus();
		boolean odef = oldstatus == Flags.DEFINED;
		
		if (def && odef && (!newsym.isFunction() || 
				            ((Function)newsym).isInline()))
			error("redefinition of " + newsym.getQuotName());
    		
		Order ord = TypeEquiv.cmpByName(newsym.getType(), oldsym.getType());

		if (ord == Order.INCOMPATIBLE) 
			error("conflicting types for " + newsym.getQuotName() + 
					"\n" + newsym.getType() + "\n" + oldsym.getType());
		
		if (ord == Order.GREATER)
			oldsym.setType(newsym.getType());
	        		
		if ((oldstatus == Flags.COMMON && status == Flags.DEFINED) ||
		    (oldstatus == Flags.UNDEFINED && status == Flags.COMMON) ||
		    (oldstatus == Flags.UNDEFINED && status == Flags.DEFINED))
			oldsym.setStatus(status);
		
		oldsym.addStorage(newsym.getStorage());
    }
    
    private static void checkAll(Symbol sym, Symtab tab, Symtab global) 
    	throws SemanticError
    {
    	Symbol oldsym = null;
    	while (tab != null && tab != global) {
    		oldsym = tab.lookupSymbol(sym);
    		if (oldsym != null)
    			error("Shadowed declaration for " + sym.getQuotName());
    		tab = tab.getParent();
    	}
    }
    
    static void addSpecifier(Specifiers s) throws SemanticError {
    	Specifiers crt = specifiersStack.peek();
    	if (crt != null)
    		crt.merge(s);
    	else {
    		specifiersStack.pop();
    		specifiersStack.push(s);
    	}
    }
    
    static void addInitializer(Symbol sym, Initializer init) 
    	throws SemanticError 
    {
		initialize(new VarExpr(sym), init, initBlockStack.peek());
    }
    
    static private void initialize(Expression e, Initializer init, 
    		                       BlockStat block) throws SemanticError 
    {
    	Type t = e.getType().unwrap();
    	Expression ie = init.getSingleInit();

    	/* Here the compiler has to guess what the user actually meant... */
    	if (init instanceof InitExpr || 
    			(ie != null && !t.isStruct()) ) {
    		if (!t.isArray()) {
    			if (!ie.getType().unwrap().isStruct())
    				while (t.isStruct()) {
    					AggregateType at = (AggregateType)t;
    					Symbol firstfield = at.getField(0);
    					t = firstfield.getType();
    					e = new FieldExpr(false, e, firstfield);
    				}
    			block.add(new ExprStat(new AssignExpr(e, ie)));
        		return;
    		}
    		else if (ie.getType().isPointer()) {
    			/* assignment would be invalid; need to use strcpy */
    			Expression func = new VarExpr(getStrcpy());
    			List<Expression> args = new ArrayList<Expression>();
    			args.add(coerceCharPtrType(e));
    			args.add(ie);    			
    			block.add(new ExprStat(new CallExpr(func, args)));
        		return;
    		}
        }
        
    	if (!(init instanceof InitList)) {
    		error("Cannot handle initializer: " + e + " = " + init);
    	}
    	
    	InitElement elem = ((InitList)init).inits;
    	if (elem == null) return;
		elem = elem.fillInit(t);

		for (; elem != null; elem = elem.next) {
			Expression e1 = elem.getElement(e);
			initialize(e1, elem.val, block);
		}
    }
    
	private static Expression coerceCharPtrType(Expression e) {
		Type type = e.getType().unwrap();
		Type tgt = type.getTargetType();
		return tgt != null && tgt.isPrimitive("char") 
			? e :  new CastExpr(Type.CharPtr, e);
	}

	private static Symbol getStrcpy() {
		Symtab current = Symtab.getCurrentScope();
		Symbol sym = current.lookupSymbolRecurse(Flags.FUN, "strcpy");
		if (sym == null) {
			FunType ftype = new FunType(Type.CharPtr);
			ftype.addArg(Type.CharPtr);
			ftype.addArg(Type.CharPtr);
			sym = new Function("strcpy", ftype, Flags.UNDEFINED, 0, Loc.UNK);
			Symtab.getGlobalScope().insertSymLocal(sym);
		}
		sym.setAccessed();
		return sym;
	}

	/** Push a function scope. */
    static void pushFunScope(Declarator d, boolean oldstyle) 
                throws SemanticError 
    {    	
    	FunDeclarator fd = getFunDeclarator(d);        
        if (fd == null)
        	error("incorrect function declarator.");

    	if (oldstyle) 
    		oldStyleParams = fd.symtab;

        Symtab.pushScope(fd.symtab);
        structScopeStack.push(null);
    }

    private static FunDeclarator getFunDeclarator(Declarator d) {
        FunDeclarator fd = null;
        for (Declarator crt = d; crt != null; crt = crt.base)
        	if (crt != null &&
        		crt instanceof FunDeclarator &&
        		crt.base instanceof IdDeclarator) {
        		fd = (FunDeclarator)crt; 
        		break;
        	}
        return fd;
    }
    
    /** Fixup function symbol after an old style param declaration. */
    static void setFunSymbol(Declarator d, boolean hasspec) 
                throws SemanticError 
    {
    	oldStyleParams = null;

    	/* Assume signed int if nothing specified. */
        Specifiers s;
        if (hasspec) 
        	s = specifiersStack.pop();
        else
        	s = new Specifiers(Type.SignedInt);

        /* build the function symbol */
        Function sym = (Function) d.getSymbol(s, Flags.DEFINED);

        Symtab global = Symtab.getGlobalScope();
        Symtab current = Symtab.getCurrentScope().getParent();
        boolean declaredExtern = (s.flags & Flags.EXTERN) != 0;
        boolean declaredStatic = (s.flags & Flags.STATIC) != 0;
        boolean declaredTypedef = (s.flags & Flags.TYPEDEF) != 0;
    	boolean defaultExtern = !declaredStatic;
        boolean isExtern = declaredExtern || defaultExtern; 

    	if (declaredExtern && declaredStatic)
    		error("declaration is both extern and static");

    	if (declaredTypedef)
    		error("function cannot be declared typedef");

    	Symbol oldsym = current.lookupSymbol(sym);
    	if (oldsym != null && oldsym.isStatic()) {
    		if (declaredExtern)
    			error("function previously declared static");
    		else if (!declaredStatic)
    			isExtern = false;
    	}    			

    	Symtab thetab = null;
    	if (oldsym == null) {
    		thetab = isExtern ? global : current;

    		/* check if the extern "skips over" other declarations */
    		if (isExtern) 
    			checkAll(sym, current, global);

    		oldsym = thetab.lookupSymbol(sym);
    	}
    	
        if (oldsym == null) {
	        thetab.insertSymLocal(sym);
        }
        else {
            /* check function re-definitions */
        	multipleDeclarations(sym, oldsym);
        	Function fun = (Function)oldsym;
        	fun.setInfoFrom(sym);
        	sym = fun;
        }

        FunDeclarator fd = getFunDeclarator(d);
        fd.symtab.setFunctionSymbol(sym);
        if (fd.symtab.hasVoidArgs())
        	((FunType)sym.getType()).setVoidArgs();
        
        assert structScopeStack.peek() == null;
        structScopeStack.pop();
    }
    
    /** Return any existing label with this name, or add a new one, 
     *  also considering local label scope extension */
    static Label lookupOrAddLabel(String name) {
        Label sym;

        Symtab current = Symtab.getCurrentScope();
        if (current.isGnuLocalLabel(name))
            sym = current.lookupLabelLocal(name);
        else
            sym = Symtab.lookupLabelRecurse(name);

        if (sym != null)
            return sym;
        else
            return Symtab.addLabel(name);
    }

    static LabelStat makeLabelStatement(String id, Statement stmt) 
    	throws SemanticError {
    	Label labelSym = Actions.lookupOrAddLabel(id);
    	if (labelSym.isDefined())
    		error("Multiple definitions of label " + id);
       
    	LabelStat labelStat = new LabelStat(labelSym, stmt);
    	labelSym.setLabelStat(labelStat);
    	return labelStat;
    }
    
	public static void markLocalLabel(String id) {
		HashSymtab tab = (HashSymtab)Symtab.getCurrentScope();
		tab.markGnuLocalLabel(id);
	}

    private static class GotoChecker extends SimplePass {
    	/* Queue up the error as we can't throw it through the visitor */
    	SemanticError error; 
    	
        void checkGotos(BlockStat body) throws SemanticError {
            process(body);
            if (error != null) throw error;
        }

        protected void localVisitPre (GotoStat e) {
            Label ls = e.getLabel();
            if (ls.isUndefined())
               error = new SemanticError("use of undefined label " + 
               		                           ls.getQuotName());
        }
    }

    /** Extract the type of the specifier on top of the
     * specifiers stack. Combine it with declarator d, if non-null. */
    static Type getType(Declarator d) throws SemanticError {
    	Specifiers s = specifiersStack.pop();
    	Type t = s.getType();    	
    	if (d == null) return t;
    	else return d.getType(t);
    }
    
    /** Pop a function scope. */
    static void popFunScope(BlockStat body) throws SemanticError {
    	Symtab st = Symtab.getCurrentScope();
    	if (!(st instanceof FunSymtab)) 
    		error("expected function symtab");

    	FunSymtab fst = (FunSymtab)st;
    	Function fsym = fst.getFunctionSymbol();
    	(new GotoChecker()).checkGotos(body);
    	fsym.setBody(body);
    	fsym.setParams(fst.getParams());
    	Symtab.popScope();

    	if (fsym.getName().equals("main")) 
    		mainfun = fsym;
    }
    
    /** Push a new scope for parsing parameters. */
    static void pushParamScope() {
    	Symtab.pushScopeNoLink(new FunSymtab());
    	structScopeStack.push(null);
    }
    
    /** Pop the parameter scope. */
    static Declarator popParamScope(Declarator d) {
        FunSymtab fs = (FunSymtab)Symtab.popScope();
        assert structScopeStack.peek() == null;
        structScopeStack.pop();
        if (d == null) d = new IdDeclarator();
        return new FunDeclarator(d, fs);
    }

    /** Push a statement block scope. */
    static void pushBlockScope() {
    	HashSymtab newscope = new HashSymtab();
    	Symtab.pushScope(newscope);
        structScopeStack.push(newscope);
        initBlockStack.push(new BlockStat(newscope));
    }
    
    /** Pop a statement block scope. */
    static BlockStat popBlockScope() {
        Symtab.popScope();
        structScopeStack.pop();
    	return initBlockStack.pop();
    }
    
    static void addStatement(Statement s) {
    	initBlockStack.peek().add(s);
    }
    
    /** Push a scope for declarations in a struct, union, or enum. */
    static void pushStructScope(byte kind, String id) 
                throws SemanticError 
    {
    	boolean unnamed = (id == null);
    	if (unnamed)
    		id = Options.ANONYM_STRUCT_PREFIX + (c++);
    	
    	Symtab scope = structScopeStack.peek();
    	if (scope == null)
        	error("struct definition in parameter list");
        
        Symbol sym = scope.lookupSUELocal(id);
        
        if (sym != null) {
        	if (sym.getKind() != kind)
        		error("incorrect kind for " + sym.getQuotName());
            
        	if (sym.getType() != null)
        		error("redefinition of " + sym.getQuotName());
        } 
        else {
            sym = Symbol.newStruct(id, kind);
            if (unnamed) sym.setAnonymous();
            scope.insertSymLocal(sym);
        }

        StructSymtab symtab = new StructSymtab(kind);
        symtab.tag = sym;
        Symtab.pushScopeNoLink(symtab);
        /* no pushing on structScopeStack! */
    }

    /** Pop a struct/union/enum scope and create its type. */
    static Type popStructScope(int kind) throws SemanticError {
        /* no popping off structScopeStack! */
    	Symtab scope = structScopeStack.peek();
    	StructSymtab stab = (StructSymtab)Symtab.popScope(); 
    	AggregateType type = new AggregateType(stab.kind);
        for (Symbol s : stab.getVFTSymbols()) 
            if (s.getKind() == Flags.VAR) 
                type.addField(s);
            else 
            	error("invalid declaration in struct.");
        
        Symbol sym = stab.tag;
        if (sym.getType() != null) 
            error("redefinition of " + sym.getQuotName());
        
        sym.setType(type);
        sym.setStatus(Flags.DEFINED);
        
        if (type.isEnum()) {
			Type ent = new AggrNameType(sym);
			for (int i = 0; i < type.numFields(); i++) {
				Symbol enumsym = type.getField(i);
				Symbol oldsym = scope.lookupVFTLocal(enumsym.getName());
				if (oldsym != null)
					error("redefinition of " + enumsym.getQuotName());

				enumsym.setType(ent);
				scope.insertSymLocal(enumsym);
			}
        }

        return new AggrNameType(sym);
    }

    /** Enter a struct/union/enum name in current scope. */
    static Type enterStructName(byte k, String id) throws SemanticError {
        Symbol sym = Symtab.getCurrentScope().lookupSUERecurse(id);
        if (sym != null) {
            if (sym.getKind() != k)
                error("incorrect kind for " + sym.getQuotName());
        } else {
            sym = Symbol.newStruct(id, k);
            Symtab scope = structScopeStack.peek();
            if (scope == null)
            	error("struct declaration in parameter list");
            
            scope.insertSymLocal(sym);
        }

        return new AggrNameType(sym);
    }

    /** Enter an enumeration symbol in the current enum scope. */
    static void enterEnumId(String id, Expression e, Loc loc) 
    	throws SemanticError 
    {
        Symbol sym = new EnumConstant(id, e, loc);
        Symbol oldsym = Symtab.getCurrentScope().lookupSymbol(sym);
        
        if (oldsym != null) 
            error("multiple occurrences of " + id );

        Symtab.getCurrentScope().insertSymLocal(sym);
    }
    
    /** Adds a nested struct with no declarator. */
    static void processNestedStruct() throws SemanticError {
    	Type t = specifiersStack.peek().getType();
    	if (!(t instanceof AggrNameType))
    		error("fields with no declarators must be structs.");

    	Actions.processDeclaration(new IdDeclarator(), false, true);
   }
    
    /** Creates an empty global void -> void function with given name, 
     * and returns the symbol */
    static private Function makeInitFunction(String name) 
    	throws SemanticError 
   {
        /* Create function scope. */
        FunSymtab fst = new FunSymtab();
        Symtab.pushScope(fst);
        
        /* ... and the body scope */
        HashSymtab tab = new HashSymtab();
		Symtab.pushScope(tab);
        
        /* Create symbol */
        Function fs = Function.initFun(name, Lexer.getFile());
        fst.setFunctionSymbol(fs);
                
        /* Exit function body scope. */
        Symtab.popScope();
        
        /* Exit function scope */
        popFunScope(new BlockStat(tab));
                
        /* insert the symbol, register the function */
        Symtab.getGlobalScope().insertSymLocal(fs);
        
        return fs;
    }

    /** Helper function: make sure that the name is a valid ID */
    static String escapeName(String name) {
        StringBuffer out = new StringBuffer();
        for (int i = 0; i < name.length(); i++)
            if (Character.isLetter(name.charAt(i)) ||
            	Character.isDigit(name.charAt(i)))
                out.append(name.charAt(i));
            else
                out.append('_');
        return out.toString();
    }
    
    /** Push a file scope. */
    static void pushFileScope() throws SemanticError {
    	Symtab.setGlobalScope(new HashSymtab());

        /* Do the actual scope-pushing */
    	FileSymtab filetab = new FileSymtab(Lexer.getFile());
        Symtab.pushScope(filetab);
        structScopeStack.push(filetab);
    
        /* Create a function for global file-level initializations */
        String name = escapeName(Options.FILE_STARTUP_PREFIX + Lexer.getFile());
        Function initFn = makeInitFunction(name);
        initBlockStack.push(initFn.getBody());
        inits.add(initFn);
    }
    
    /** Pop a file scope. */
    static void popFileScope() throws SemanticError {
    	assert specifiersStack.isEmpty();
    	
        final HashMap<Symbol,Symbol> map = new HashMap<Symbol,Symbol>();
        HashSymtab gtab = Symtab.getGlobalScope();
    	FileSymtab filetab = (FileSymtab)Symtab.popScope();
    	structScopeStack.pop();
        initBlockStack.pop();

        /* Remove the initializer if it is empty */
        int last = inits.size() - 1;
        Function initFn = inits.get(last);
        if (initFn.getBody().numStatKids() == 0) {
        	initFn.getSymTab().getParent().detach();
        	inits.remove(last);
        	Symtab.getGlobalScope().removeSymLocal(initFn);
        }
        
        /* collect functions defined in this file */
        Set<Function> filefuncs = new HashSet<Function>();
    	for (Symbol sym : gtab.getVFTSymbols()) {
        	if (!sym.isFunction()) continue;
        	Function fun = (Function)sym;        	
        	if (fun.isDefined()) filefuncs.add(fun);
    	}        
    	for (Symbol sym : filetab.getVFTSymbols()) {
        	if (!sym.isFunction()) continue;
        	Function fun = (Function)sym;
        	if (fun.isDefined()) filefuncs.add(fun);
    	}        

    	/* move var/fun symbols from the file scope to allglobal.
    	 * Extern symbols that are not used in this file are not moved. */
        if (allglobal == null) 
        	allglobal = new HashSymtab();

        Iterator<Map.Entry<String, Symbol>> it;
        it = gtab.getVFTMap().entrySet().iterator();
        while (it.hasNext()) {
        	Map.Entry<String, Symbol> e = it.next();
        	Symbol sym = e.getValue();
        	
        	if (!sym.isUndefined() || sym.isAccessed()) {
            	it.remove();
        		moveSymbol(sym, allglobal, map, filefuncs);
        	}
        }
        
		/* move typdefs, struct/union/enum names, and enum constants */
        it = filetab.getSUEMap().entrySet().iterator();
        while (it.hasNext()) {
        	Map.Entry<String, Symbol> e = it.next();
        	Symbol sym = e.getValue();
        	it.remove();
    		moveSymbol(sym, allglobal, map, filefuncs);
        }

        it = filetab.getVFTMap().entrySet().iterator();
        while (it.hasNext()) {
        	Map.Entry<String, Symbol> e = it.next();
        	Symbol sym = e.getValue();
        	if (!sym.isStatic() && 
        			(sym.getKind() == Flags.TYPE || 
        			 sym instanceof EnumConstant)) {
            	it.remove();
        		moveSymbol(sym, allglobal, map, filefuncs);
        	}
        }
                
        /* fix references in symbol tables and in the AST. */
        allglobal.addKid(filetab);
        fixupSymtab(allglobal, map);
        for (Function fun : filefuncs)
    		fixupAST(fun.getBody(), fun.getSymTab(), map);
        
        /* cleanup */
        Symtab.setGlobalScope(null);
        
        Symtab.recordFileFunctions(filetab.getFileName(), filefuncs);
    }
    
	private static void moveSymbol(Symbol fsym, HashSymtab tab, 
    							   Map<Symbol,Symbol> map,
    							   Set<Function> funset) 
    		throws SemanticError 
   {
    	Symbol gsym = tab.lookupSymbol(fsym);
    	if (gsym == null) {
    		if (Flags.isSUETag(fsym.getKind())) 
    			for (Symbol sym : tab.getSUESymbols()) 
    				if (sym.isAnonymous()) {
    					Type ftype = fsym.getType();
    					Type type = sym.getType();
    					if (TypeEquiv.cmpByStruct(ftype, type) == Order.EQUAL) {
    						gsym = sym;
    						break;
    					}
    				}
    		if (gsym == null) {
    			allglobal.insertSymLocal(fsym);
    			return;
    		}
    	}
    	
		map.put(fsym, gsym);
		if (funset.contains(fsym)) {
			funset.remove(fsym);
			funset.add((Function)gsym);
		}

		if (gsym.getType() == null) 
    		gsym.setType(fsym.getType());
    	else {
    		boolean fdef = fsym.isDefined();
    		boolean gdef = gsym.isDefined();
    		byte kind = fsym.getKind();
    		boolean isvar = kind == Flags.VAR && 
    		                !(fsym instanceof EnumConstant);
    		boolean isfun = fsym.isFunction();
    		
    		if (fdef && gdef && (isvar || isfun)) {
    			if (isfun && fsym.isInline()) {
    				/* ignore new defs of inline functions */
    				return;
    			}
    			error(fsym.getQuotName() + " defined in multiple files");
    		}
    		
    		Order ord;
    		if (fsym.isFunction() && gsym.isFunction()) {
    			if (fdef && !gdef) ord = Order.GREATER;
    			else if (!fdef && gdef) ord = Order.LESS;
    			else ord = Order.NONE;
    		} else
        		ord = TypeEquiv.cmpByStruct(fsym.getType(), gsym.getType());
        		
    		if (ord == Order.INCOMPATIBLE) {
	    		/* turn off warnings for enum constants */
	    		if (fsym instanceof EnumConstant) {
	    			assert gsym instanceof EnumConstant;
	    			fsym.setFreshName(allglobal);
		    		allglobal.insertSymLocal(fsym);
		    		return;
	    		}
	    		else {
    	    		String oldname = gsym.getName();
    	    		String filename = gsym.getLoc().toString();
    	    		String newfilename = fsym.getLoc().toString();
    	    		TypeEquiv.cmpByStruct(fsym.getType(), gsym.getType());
    				error("\n\tIncompatible types for " + 
    						Flags.kindToString(gsym.getKind()) + 
    						" '" + oldname + "' at: '" + newfilename + "'" +
    						"\n\tPrevious declaration at: '" + filename + "'");
	    		}
    		}
    		
    		if (ord == Order.GREATER) {
    			if (gsym.isFunction()) 
        			((Function)gsym).setInfoFrom((Function)fsym);
    			else if (!gsym.isTag())
    				gsym.setType(fsym.getType());
    		}
    		
    		byte gstatus = gsym.getStatus();
    		byte fstatus = fsym.getStatus();
    		if ((gstatus == Flags.COMMON && fstatus == Flags.DEFINED) ||
        		(gstatus == Flags.UNDEFINED && fstatus == Flags.COMMON) ||
        		(gstatus == Flags.UNDEFINED && fstatus == Flags.DEFINED)) 
    		{
        		gsym.setStatus(fstatus);
        		gsym.setLoc(fsym.getLoc());
    		}

    		gsym.addStorage(fsym.getStorage());
    	}
    }

    private static void fixupAST(Statement stat, Symtab tab, 
    		Map<Symbol,Symbol> map) 
    {
    	if (stat == null) return;
	    	
    	if (stat instanceof BlockStat) 
    		tab = ((BlockStat)stat).getSymTab();
	    		
    	for (int i = 0; i < stat.numStatKids(); i++)
    		fixupAST(stat.getStatKid(i), tab, map);
			
    	for (int i = 0; i < stat.numExprKids(); i++)
    		fixupAST(stat.getExprKid(i), tab, map);
    }

    private static void fixupAST(Expression expr, Symtab tab, 
    		Map<Symbol,Symbol> map) 
    {
    	if (expr == null) return;
			
    	for (int i = 0; i < expr.numExprKids(); i++)
    		fixupAST(expr.getExprKid(i), tab, map);

    	for (int i = 0; i < expr.numTypeKids(); i++)
    		fixupType(expr.getTypeKid(i), tab, map);

    	for (int i = 0; i < expr.numStatKids(); i++)
    		fixupAST(expr.getStatKid(i), tab, map);
    		
    	if (expr instanceof VarExpr) {
    		VarExpr e = (VarExpr)expr;
    		Symbol oldvar = e.getSymbol();
    		Symbol newvar = map.get(oldvar);
    		if (newvar != null && newvar != oldvar) 
    			e.setSymbol(newvar);
		}
    	else if (expr instanceof FieldExpr) {
    		FieldExpr e = (FieldExpr)expr;
    		Symbol oldfield = e.getField();
    		Type etype = e.getBase().getType().unwrap();
    		if (e.isDeref()) 
    			etype = etype.getTargetType().unwrap();
    		AggregateType atype = (AggregateType)etype;
    		Symbol newfield = atype.getField(oldfield.getName());
    		assert newfield != null;
    		if (newfield != oldfield)
    			e.setField(newfield);
    	}
    }

    private static void fixupSymtab(Symtab tab, Map<Symbol, Symbol> map) {
    	for(Symbol sym : tab.getAllSymbols()) 
    		fixupType(sym.getType(), tab, map);

    	for (Symtab ktab : tab.getKids())
    		fixupSymtab(ktab, map);
    }

    private static void fixupType(Type t, Symtab tab, Map<Symbol, Symbol> map) 
    {
    	if (t == null) return;
    	    	
    	if (t instanceof TypedefType) {
    		TypedefType tt = (TypedefType)t;
    		Symbol sym = tt.getTypeSymbol();
    		Symbol newsym = map.get(sym);
    		if (newsym != null && newsym != sym) 
    			tt.setTypeSymbol(newsym);
    	} 
    	else if (t instanceof AggrNameType) {
    		AggrNameType st = (AggrNameType)t;
    		Symbol sym = st.getTypeSymbol();
    		Symbol newsym = map.get(sym);
    		if (newsym != null && newsym != sym)
    			st.setTypeSymbol(newsym);    		
    	} 
    	else {
    		if (t instanceof ArrayType) {
    			ArrayType at = (ArrayType)t;
    			Expression e = at.getExpression();
    			if (e != null)
    				fixupAST(e, tab, map);
    		}
    			
    		for (int i = 0; i < t.numTypeKids(); i++)
    			fixupType(t.getTypeKid(i), tab, map);
    	}
   	}
    
    /** Build an expression representing __FUNCTION__ */
    static StringExpr buildFunctionName() {
    	Symtab tab = Symtab.getCurrentScope();
    	FileSymtab filetab = tab.getEnclosingFunctionScope(); 
    	return new StringExpr(filetab.getFileName());
    }
    
    /** Build a VarExpr from an identifier name. Requires that
     *  undeclared symbols are function symbols. */
    static VarExpr buildVarExpr(String id) throws SemanticError {
    	id = Builtins.stripBuiltinPrefix(id);
		Symbol sym = Symtab.getCurrentScope().lookupVFTRecurse(id);
		if 	(sym == null) {
			sym = Function.defaultFun(id);
			Symtab.getGlobalScope().insertSymLocal(sym);
		}
		sym.setAccessed();
		return new VarExpr(sym);
    }

    /** Build a CallExpression. Insert function symbol into the
     *  global symtab, if not present in scope. */
    static CallExpr buildCallExpr(Expression fun, List<Expression> args) {
    	if (args == null)
    		args = new ArrayList<Expression>();
    	
    	/* If a funtion symbol is used without being declared,
    	 * set its type according to this call, and enter it in the 
    	 * symbol table. */
    	if (fun instanceof VarExpr) {
    		VarExpr ve = (VarExpr)fun;
    		Symbol sym = ve.getSymbol();
    		if (sym.isFunction()) {
    			Function fsym = (Function)sym;
    			if (fsym.getStatus() == Flags.NONE) {
    				sym.setStatus(Flags.UNDEFINED);
    			}
    		}
    	}
    	
    	return new CallExpr(fun, args);
    }
    
    static Expression buildCommaExpr(List<Expression> v) {
    	if (v.size() == 1)
    		return v.get(0);
    	else
    		return new CommaExpr(v);
    }

    /** Checks if all symbols in the expression are in scope. */
    static Expression checkSymbols(Expression e) throws SemanticError {
    	if (e == null) return null;
    	
    	if (e instanceof VarExpr) {
    		VarExpr ve = (VarExpr)e;
    		Symbol sym = ve.getSymbol();
    		if (sym.getStatus() == Flags.NONE)
    			error("Undeclared symbol " + sym.getQuotName());
    	}
    	
    	for (int i = 0; i < e.numExprKids(); i++) 
    		checkSymbols(e.getExprKid(i));
    	return  e;
    }
    
    /** Build a FieldExpr from an expression and a field name. */
    static FieldExpr buildFieldExpr(boolean deref, Expression exp, String id) 
    				 throws SemanticError {
    	Type t = ExprType.typeOf(exp);
    	if (deref) {
    		t = t.unwrap();
    		if (t instanceof PointerType)
    			t = ((PointerType)t).getTargetType();
    		else if (t instanceof ArrayType)
    			t = ((ArrayType)t).getElementType();
    		else
    			error("Non-pointer type in field dereference.");
    	}

		Type real = t.unwrap();
    	if (real == null) {
    		error("Field access for an incomplete type" + 
    				"\n\t field: " + id + 
    				"\n\t type : " + t.toString() +
    				"\n\t expr : " + ASTPrinter.toString(exp, false));
    	}

    	if (!(real instanceof AggregateType))
    		error("Field access for a non-struct type: field " + id);

    	AggregateType struct = (AggregateType)real;
    	FieldExpr fexpr = struct.buildFieldExpr(deref, exp, id);
    	if (fexpr == null) 
    		error("Invalid field " + id);
    	return fexpr;
    }

	public static Expression buildCompoundLiteral(Type t, InitElement e) 
		throws SemanticError 
	{
		pushBlockScope();
		Symbol temp = Symbol.newTemp("__t", t);
		Symtab.getCurrentScope().insertSymLocal(temp);
		addInitializer(temp, new InitList(e));
		initBlockStack.peek().add(new ExprStat(new VarExpr(temp)));
		return new StatExpr(Actions.popBlockScope());
	}

	public static void startParsing() {
		allglobal = new HashSymtab();
		Symtab.setGlobalScope(allglobal);
		structScopeStack.push(allglobal);
		Builtins.registerBuiltins(); 
	}

	public static void doneParsing() {
		Symtab.setGlobalScope(allglobal);
		assert allglobal != null;
		
		if (mainfun != null)
			for (Function initFn : inits) {
				CallExpr call = new CallExpr(new VarExpr(initFn));
				mainfun.getBody().addFirst(new ExprStat(call));
				
				for (Symbol sym : initFn.getSymTab().getVFTSymbols()) 
					if (sym.hasAddressTaken())
						sym.addStorage(Flags.STATIC);
			}
		
        if (SANITY_CHECKS) 
        	SanityChecks.run();
	}
	
	static void error(String msg) throws SemanticError {
		System.out.flush();
		throw new SemanticError(msg);
	}
}
