package crystal.parser;

import crystal.ast.*;
import crystal.general.*;

abstract class Initializer {
	abstract Expression getSingleInit();
}

class InitExpr extends Initializer {
    Expression expr;
    
    InitExpr(Expression e) {
    	expr = e;
    }
    
    public String toString() {
    	return expr.toString();
    }

	Expression getSingleInit() {
		return expr;
	}
}

class InitList extends Initializer {
    InitElement inits;
    
    InitList(InitElement l) {
    	inits = l == null ? null : l.reverse();
    }

	public String toString() {
		return "(" + inits + ")";
	}

	Expression getSingleInit() {
		if (inits.next == null) 
			return inits.val.getSingleInit();
		else
			return null;
	}
}

class InitElement {
	Designator des;
	Initializer val;
	InitElement next;
	InitElement prev;

	
	public InitElement(Designator d, Initializer v) {
		des = d;
		val = v;
	}

	public InitElement add(InitElement n) {
		n.prev = this;
		return n;
	}

	InitElement fillInit(Type t) {
		int k = 0;
		if (!(t instanceof ArrayType || t instanceof AggregateType)) {
			assert next == null;
			return this;
		}
			
		InitElement elem = this;
		InitElement last = null;
		for (; elem != null; k++) {
			if (elem.des == null) {
				elem.des = Designator.getDefault(k, t);
				elem.prev = last;
				last = elem;
				elem = elem.next;
			}
			else {
				elem = elem.expandRange();
				if (elem.des.match(k, t)) {
					elem.prev = last;
					last = elem;
					elem = elem.next;
				}
			}
		}
		
		elem = last.reverse();
		 
		/* set array size if not mentioned */
		if (t instanceof ArrayType) {
			ArrayType at = (ArrayType)t;
			if (at.getExpression() == null) {
				int max = -1;
				for (InitElement e = elem; e != null; e = e.next) {
					int n = e.des.n1.intValue();
					if (n > max) max = n;
				}
				at.setExpression(max + 1);
			}
		}
		
		return elem;
	}
	
	InitElement reverse() {
		InitElement crt = this;
		while (crt.prev != null) {
			InitElement tmp = crt.prev;
			tmp.next = crt;
			crt = tmp;
		}
		return crt;
	}
	
	private InitElement expandRange() {
		if (des.id != null)
			return this;
		
		int m1 = des.n1.intValue();
		int m2 = des.n2.intValue();
		if (m1 == m2)
			return this;
		
		if (m1 > m2)
			SemanticError.err("invalid range in initializer: " + des);

		InitElement elem = this;
		InitElement nelem = null;
		Designator nxt = des.next;
		des.n1 = des.n2;
		for (int k = m2-1; k >= m1; k--) {
			Integer nk = new Integer(k);
			Designator ndes = new Designator(nk, nk);
			ndes.next = nxt;
			nelem = new InitElement(ndes, val);
			nelem.next = elem;
			elem = nelem;
		}
		return nelem;
	}

	Expression getElement(Expression e) {
		if (des == null)
			return e;
		
		for (; des != null; des = des.next)
			if (des.id != null) {
				AggregateType st = (AggregateType)e.getType().unwrap();
				e = st.buildFieldExpr(false, e, des.id);
			}
			else 
				e = new ArrayExpr(e, new NumberExpr(des.n1));
		
		return e;
	}
	
	public String toString() {
		String s = "";
		
		if (des != null)
			s += des.toString() + " = ";
		
		s += val.toString();

		if (next != null)
			s += ", " + next;
		
		return s;
	}
}


class Designator {
	Designator next;
	String id;
	Number n1, n2;
	
	Designator(String s) {
		id = s;
	}
	
	Designator(Expression e1, Expression e2) {
		n1 = e1.constValue();
		assert n1 != null;

		if (e2 == null) {
			n2 = n1;
		}
		else {
			n2 = e2.constValue();
			assert n2 != null;
		}
	}
	
	Designator(Number m1, Number m2) {
		n1 = m1;
		n2 = m2;
	}
	
	public static Designator getDefault(int k, Type t) {
		if (t instanceof ArrayType) {
			Integer nk = new Integer(k);
			return new Designator(nk, nk);
		}
		else if (t instanceof AggregateType) {
			AggregateType st = (AggregateType)t;
			if (k >= st.numFields())
				SemanticError.err("invalid initializer.");
			
			String str = st.getField(k).getName();
			return new Designator(str);
		}
		
		assert false;
		return null;
	}

	public boolean match(int k, Type t) {
		if (t instanceof ArrayType) {
			assert n1 != null;
			int n = n1.intValue();
			if (n < k)
				SemanticError.err("invalid initializer.");
			
			return n == k;
		}
		else if (t instanceof AggregateType) {
			AggregateType st = (AggregateType)t;
			if (k >= st.numFields())
				SemanticError.err("invalid initializer.");
			
			return st.getField(k).getName().equals(id);
		}
			
		assert false;
		return false;
	}

	Designator add(Designator n) {
		this.next = n;
		return this;
	}
	
	public String toString() {
		String s = id;
		if (s == null) 
			s = n2 != null 
				? "[" + n1 + " .. " + n2 + "]"
				: "[" + n1 + "]";
	
		if (next == null)
			return s;
		else 
			return s + next.toString();
	}
}
