package crystal.parser;

import java.util.*;

import crystal.ast.*;
import crystal.general.*;

class SanityChecks {

	private static Set<Type> checked;
	
	static void run() {
		/* check the symbol tables */
		checked = new HashSet<Type>();
		checkSymtab(Symtab.getGlobalScope());
		
		/* check the code */
        for (Function fun : Symtab.getFunctions())
    		checkAST(fun.getBody(), fun.getSymTab());
	}
	
	static void check(Symtab global, List<Function> funs) {
		checked = new HashSet<Type>();
		checkSymtab(global);
        for (Function fun : funs)
    		checkAST(fun.getBody(), fun.getSymTab());
	}
	
	private static void checkSymtab(Symtab tab) {
		for (Symbol s : tab.getAllSymbols())
			checkType(s.getType(), tab);
		
		for(Symtab kidtab : tab.getKids()) 
			checkSymtab(kidtab);
	}
	
    private static void checkType(Type t, Symtab tab) 
    {
    	if (t == null) return;
    	
    	if (checked.contains(t)) return;
    	else checked.add(t);
    	
    	if (t instanceof AggrNameType) {
    		AggrNameType agg = (AggrNameType)t;
    		Symbol sym = agg.getTypeSymbol();
    		assert sym == tab.lookupSymbol(sym) ||
    		       sym == tab.getParent().lookupSymbolRecurse(sym);
    	}
    	
    	if (t instanceof TypedefType) {
    		Symbol sym = t.getTypeSymbol();
    		assert sym.getKind() == Flags.TYPE;
    		assert sym == tab.lookupSymbol(sym) ||
    			   sym == tab.getParent().lookupSymbolRecurse(sym);
    	} 

    	for (int i = 0; i < t.numTypeKids(); i++)
    		checkType(t.getTypeKid(i), tab);
    }

    private static void checkAST(Statement stat, Symtab tab) 
    {
    	if (stat == null) return;
	    	
    	if (stat instanceof BlockStat) 
    		tab = ((BlockStat)stat).getSymTab();
	    		
    	for (int i = 0; i < stat.numStatKids(); i++)
    		checkAST(stat.getStatKid(i), tab);
			
    	for (int i = 0; i < stat.numExprKids(); i++)
    		checkAST(stat.getExprKid(i), tab);
    }

    private static void checkAST(Expression expr, Symtab tab) 
    {
    	if (expr == null) return;
			
    	for (int i = 0; i < expr.numExprKids(); i++)
    		checkAST(expr.getExprKid(i), tab);

    	for (int i = 0; i < expr.numTypeKids(); i++)
    		checkType(expr.getTypeKid(i), tab);

    	for (int i = 0; i < expr.numStatKids(); i++)
    		checkAST(expr.getStatKid(i), tab);
    		
    	if (expr instanceof VarExpr) {
    		VarExpr e = (VarExpr)expr;
    		Symbol sym = e.getSymbol();
    		assert sym == tab.lookupSymbol(sym) ||
    			   sym == tab.getParent().lookupSymbolRecurse(sym);
		}
    	else if (expr instanceof FieldExpr) {
    		FieldExpr e = (FieldExpr)expr;
    		Type ftype = e.getField().getType();
    		checkType(ftype, tab);

    		Type etype = e.getBase().getType().unwrap();
    		if (e.isDeref()) 
    			etype = etype.getTargetType().unwrap();
    		AggregateType atype = (AggregateType)etype;
    		Symbol field = atype.getField(e.getField().getName());
    		assert field == e.getField();
    	}
    }
}
