/* Simple bitvector implementation, without dynamic resizing.
 * The first 64 bits are inlined. */

package crystal.util;

import java.util.*;

public class BitVector {
	
	private long bits;
    private long[] more;

    private BitVector() {
    }
    
    public BitVector(int num) {
    	if (num > 64)
    		more = new long[1 + ((num-65) >> 6)];
    }
    
    public boolean get(int bit) {
    	assert bit >= 0;
        long mask = (1L << bit);

        if (bit < 64)
    		return (bits & mask) != 0L;
        assert more != null: bit;
        return (more[(bit-64) >> 6] & mask) != 0L;
    }
    
    public void set(int bit) {
    	assert bit >= 0;
        long mask = (1L << bit);

        if (bit < 64)
        	bits |= mask;
        else 
        	more[(bit-64) >> 6] |= mask;
    }


	public void setValue(int bit, boolean val) {
		if (val) set(bit);
		else clear(bit);
	}

	public void clear(int bit) {
    	assert bit >= 0;
        long mask = ~(1L << bit);

        if (bit < 64) 
    		bits &= mask;
    	else  	
    		more[(bit-64) >> 6] &= mask;
    }
    
    public boolean or(BitVector other) {
        boolean changed = false;
        long newbits = bits | other.bits;
        changed |= (newbits != bits);
        bits = newbits;

        long[] otherbits = other.more;
        if (otherbits == null)
        	return changed;
        
        assert more != null;
        assert otherbits.length <= more.length;
        for(int i = 0; i < otherbits.length; i++) {
        	long oldbits = more[i];
            newbits = oldbits | other.more[i];
            changed |= (newbits != oldbits);
            more[i] = newbits;
        }
        
        return changed;
    }
    
    public boolean and(BitVector other) {
        boolean changed = false;
        long newbits = bits & other.bits;
        changed |= (newbits != bits);
        bits = newbits;

        long[] otherbits = other.more;
        if (otherbits == null)
        	return changed;
        
        assert more != null;
        assert otherbits.length <= more.length;
        for(int i = 0; i < otherbits.length; i++) {
        	long oldbits = more[i];
            newbits = oldbits & other.more[i];
            changed |= (newbits != oldbits);
            more[i] = newbits;
        }
        
        return changed;
    }

    public boolean andNot(BitVector other) {
        boolean changed = false;
        long newbits = bits & ~other.bits;
        changed |= (newbits != bits);
        bits = newbits;

        long[] otherbits = other.more;
        if (otherbits == null)
        	return changed;
        
        assert more != null;
        assert otherbits.length <= more.length;
        for(int i = 0; i < otherbits.length; i++) {
        	long oldbits = more[i];
            newbits = oldbits & ~other.more[i];
            changed |= (newbits != oldbits);
            more[i] = newbits;
        }
        
        return changed;
    }
    
    public void not() {
        bits = ~bits;
        if (more != null) 
            for(int i = 0; i < more.length; i++) 
                more[i] = ~more[i];
    }
        
    public void clear() {
        bits = 0;
        if (more != null) 
            for(int i = 0; i < more.length; i++) 
                more[i] = 0;
    }
        
    public boolean equals( Object o ) {
        if( !( o instanceof BitVector ) ) 
        	return false;
        
        BitVector other = (BitVector) o;
        if (bits != other.bits)
        	return false;
        
        if (more == null) 
        	return other.more == null;
        
        if (more.length != other.more.length)
        	return false;
        
        for (int i = 0; i < more.length; i++)
        	if (more[i] != other.more[i])
        		return false;

        return true;
    }
    
    public int hashCode() {
        return (int)bits;
    }
    
    public BitVector copy() {
        BitVector v = new BitVector();
        v.bits = this.bits;
        
        if (more != null) {
        	v.more = new long[more.length];
        	for (int i = 0; i < more.length; i++)
        		v.more[i] = more[i];
        }

        return v;
    }
    
    public String toString() {
    	
        StringBuffer buf = new StringBuffer();
        buf.append('{');
        buf.append(String.format("%022o.", bits));
        if (more != null) 
        	for (int i = 0; i < more.length; i++)
        		buf.append(String.format("%022o.",more[i]));
        buf.append('}');
        return buf.toString();
    }

	public List<Integer> indexList() {
		List<Integer> v = new ArrayList<Integer>();
		if (bits != 0L) 
			addBits(bits, 0, v);
		
		if (more != null)
			for (int i = 0; i < more.length; i++)
				if (more[i] != 0L)
					addBits(more[i], (i+1)*64, v);
		
		return v;
	}
	
	private void addBits(long b, int o, List<Integer> v) {
		for (int i=0; i < 64; i++) {
			if ((b & 1L) == 1L) 
				v.add(o+i);
			b = b >> 1;
		}
	}

	public boolean isEmpty() {
		if (bits != 0L)
			return false;
		
		if (more != null)
			for (int i = 0; i < more.length; i++)
				if (more[i] != 0L)
					return false;
		
		return true;
	}

	public int size() {
		if (more == null)
			return 64;
		return 64 * (1 + more.length);
	}
}


